/* float.c   by Michael Thorpe   2017-05-29 */

#include <ctype.h>
#include <limits.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "lisp.h"

/* Define this if you want string_to_float to forbid "infinity" or "nan" */
/* #define NO_INFINITY_OR_NAN */

static obj eobj={&float_objtype,{refcount:1},{f:M_E}};
static obj piobj={&float_objtype,{refcount:1},{f:M_PI}};

/* Takes an integer and upconverts it to a float */
static obj *promote(obj *in) {
	obj *out;

	out=new_object(&float_objtype);
	if(out)
		out->value.f=in->value.i;
	DECREF(in);
	return(out);
}

/* Takes a newly-minted float and downconverts it to an integer if possible */
static obj *reduce(obj *in) {
	if(in->value.f<LONG_MIN || LONG_MAX<in->value.f)
		return(in);
	if(ceil(in->value.f) != in->value.f)
		return(in);
	in->type=&integer_objtype;
	in->value.i=in->value.f;
	return(in);
}

static obj *f_func(obj *a,double func(double)) {
	obj *r;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a)
			return(0);
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==a->type) {
		a=bigint_to_float(a);
		if(!a)
			return(0);
	}
#endif
	if(a->type != &float_objtype)
		return(throwtypeerror(a));
	r=new_object(&float_objtype);
	if(!r) {
		DECREF(a);
		return(0);
	}
	r->value.f=func(a->value.f);
	DECREF(a);
	return(r);
}

static obj *f_intfunc(obj *a,double func(double)) {
	obj *r;

	if(a->type != &float_objtype)
		return(throwtypeerror(a));
	if(a->value.f<LONG_MIN || LONG_MAX<a->value.f)
		return(throwrangecheck(a));
	r=new_object(&integer_objtype);
	if(!r) {
		DECREF(a);
		return(0);
	}
	r->value.i=func(a->value.f);
	DECREF(a);
	return(r);
}

obj *addf(obj *a,obj *b) {
	obj *r;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==a->type) {
		a=bigint_to_float(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#endif
	if(a->type != &float_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(&integer_objtype==b->type) {
		b=promote(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==b->type) {
		b=bigint_to_float(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#endif
	if(b->type != &float_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	r=new_object(&float_objtype);
	if(!r) {
		DECREF(a);
		DECREF(b);
		return(0);
	}
	r->value.f=a->value.f+b->value.f;
	DECREF(a);
	DECREF(b);
	return(reduce(r));
}

obj *div2f(obj *a,obj *b) {
	obj *r;

	if(&integer_objtype==b->type) {
#ifdef NO_FDIV_BY_0
		if(0==b->value.i) {
			DECREF(b);
			return(throw(INCREF(divideby0error),a));
		}
#endif
		b=promote(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==b->type) {
		b=bigint_to_float(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#endif
	if(b->type != &float_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
#ifdef NO_FDIV_BY_0
	if(0==b->value.f) {
		DECREF(b);
		return(throw(INCREF(divideby0error),a));
	}
#endif
	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==a->type) {
		a=bigint_to_float(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#endif
	if(a->type != &float_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	r=new_object(&float_objtype);
	if(!r) {
		DECREF(a);
		DECREF(b);
		return(0);
	}
	r->value.f=a->value.f/b->value.f;
	DECREF(a);
	DECREF(b);
	return(reduce(r));
}

static obj *ffabs(obj *a) {
	return(f_func(a,fabs));
}

static obj *facos(obj *a) {
	return(f_func(a,acos));
}

static obj *fasin(obj *a) {
	return(f_func(a,asin));
}

obj *fatan2(obj *a,obj *b) {
	obj *r;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==a->type) {
		a=bigint_to_float(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#endif
	if(a->type != &float_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(&integer_objtype==b->type) {
		b=promote(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==b->type) {
		b=bigint_to_float(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#endif
	if(b->type != &float_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	r=new_object(&float_objtype);
	if(!r) {
		DECREF(a);
		DECREF(b);
		return(0);
	}
	r->value.f=atan2(a->value.f,b->value.f);
	DECREF(a);
	DECREF(b);
	return(reduce(r));
}

static obj *fceiling(obj *a) {
	return(f_intfunc(a,ceil));
}

static obj *fcos(obj *a) {
	return(f_func(a,cos));
}

static obj *fexp(obj *a) {
	return(f_func(a,exp));
}

static obj *ffloor(obj *a) {
	return(f_intfunc(a,floor));
}

static obj *flog(obj *a) {
	return(f_func(a,log));
}

static obj *fround(obj *a) {
	return(f_intfunc(a,round));
}

static obj *fsin(obj *a) {
	return(f_func(a,sin));
}

static obj *fsqrt(obj *a) {
	return(f_func(a,sqrt));
}

static obj *ftan(obj *a) {
	return(f_func(a,tan));
}

obj *mulf(obj *a,obj *b) {
	obj *r;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==a->type) {
		a=bigint_to_float(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#endif
	if(a->type != &float_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(&integer_objtype==b->type) {
		b=promote(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==b->type) {
		b=bigint_to_float(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#endif
	if(b->type != &float_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	r=new_object(&float_objtype);
	if(!r) {
		DECREF(a);
		DECREF(b);
		return(0);
	}
	r->value.f=a->value.f*b->value.f;
	DECREF(a);
	DECREF(b);
	return(reduce(r));
}

#define NUMCMP_LT 1
#define NUMCMP_EQ 2
#define NUMCMP_GT 4
obj *numcmpf(obj *a,obj *b,int desired) {
	int cmp;
	obj *o;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==a->type) {
		a=bigint_to_float(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#endif
	if(a->type != &float_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(&integer_objtype==b->type) {
		b=promote(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==b->type) {
		b=bigint_to_float(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#endif
	if(b->type != &float_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	cmp=0;
	if(a->value.f == b->value.f)
		cmp=NUMCMP_EQ;
	if(a->value.f < b->value.f)
		cmp=NUMCMP_LT;
	if(a->value.f > b->value.f)
		cmp=NUMCMP_GT;
	if(!cmp) {
		o=cons(a,b);
		if(o)
			return(throwrangecheck(o));
		return(o);
	}
	DECREF(a);
	DECREF(b);
	if(cmp&desired)
		return(INCREF(&trueobj));
	else
		return(INCREF(&nilobj));
}

obj *string_to_float(obj *str) {
	obj *o;
	char *buf,*end;
	double f;
#ifdef NO_INFINITY_OR_NAN
	int i;
#endif

	if(str->type != &string_objtype)
		return(throwtypeerror(str));
	if(0==str->value.s.len)
		return(throwtypeerror(str));
#ifdef NO_INFINITY_OR_NAN
	i=0;
	if('+'==str->value.s.data[i] || '-'==str->value.s.data[i]) {
		i++;
		if(i==str->value.s.len)
			return(throwtypeerror(str));
	}
	if('.'==str->value.s.data[i]) {
		i++;
		if(i==str->value.s.len)
			return(throwtypeerror(str));
	}
	if(!isdigit(str->value.s.data[i]))
		return(throwtypeerror(str));
#endif
	buf=(char *)malloc(str->value.s.len+1);
	if(!buf) {
		DECREF(str);
		return(throwoom());
	}
	memcpy(buf,str->value.s.data,str->value.s.len);
	buf[str->value.s.len]='\0';
	f=strtod(buf,&end);
	if(*end) {
		free(buf);
		return(throwtypeerror(str));
	}
	free(buf);
	DECREF(str);
	o=new_object(&float_objtype);
	if(o)
		o->value.f=f;
	return(o);
}

obj *subf(obj *a,obj *b) {
	obj *r;

	if(&integer_objtype==a->type) {
		a=promote(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==a->type) {
		a=bigint_to_float(a);
		if(!a) {
			DECREF(b);
			return(0);
		}
	}
#endif
	if(a->type != &float_objtype) {
		DECREF(b);
		return(throwtypeerror(a));
	}
	if(&integer_objtype==b->type) {
		b=promote(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#ifdef USE_BIGINT
	if(&bigint_objtype==b->type) {
		b=bigint_to_float(b);
		if(!b) {
			DECREF(a);
			return(0);
		}
	}
#endif
	if(b->type != &float_objtype) {
		DECREF(a);
		return(throwtypeerror(b));
	}
	r=new_object(&float_objtype);
	if(!r) {
		DECREF(a);
		DECREF(b);
		return(0);
	}
	r->value.f=a->value.f-b->value.f;
	DECREF(a);
	DECREF(b);
	return(reduce(r));
}

static obj *float_tostring(obj *in) {
	char buf[30];
	int i;

	i=snprintf(buf,sizeof(buf),"%.15g",in->value.f);
	if(i<=0)
		return(throwrangecheck(in));
	DECREF(in);
	return(new_string(buf,i));
}

static int float_init() {
	if(store_builtin_ns1("acos",facos)
	|| store_builtin_ns1("asin",fasin)
	|| store_builtin_ns2("atan",fatan2)
	|| store_builtin_ns1("ceiling",fceiling)
	|| store_builtin_ns1("cos",fcos)
	|| store_object("#e",INCREF(&eobj))
	|| store_builtin_ns1("exp",fexp)
	|| store_builtin_ns1("abs",ffabs)
	|| store_builtin_ns1("floor",ffloor)
	|| store_builtin_ns1("log",flog)
	|| store_object("#pi",INCREF(&piobj))
	|| store_builtin_ns1("round",fround)
	|| store_builtin_ns1("sin",fsin)
	|| store_builtin_ns1("sqrt",fsqrt)
	|| store_builtin_ns1("string->float",string_to_float)
	|| store_builtin_ns1("tan",ftan))
		return(1);
	return(0);
}

objtype float_objtype={
	name: "float",
	tostring: float_tostring,
	init: float_init,
};

